
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <avr/io.h>
#include <avr/interrupt.h>

#include <util/twi.h>	
#include <util/delay.h> 


/** -------------------- TEMP,HUMI -------------------- **/
typedef union 
{ unsigned int i;
  double f;
} value;
enum {TEMP,HUMI};

/** -------------------- DEWPOINT --------------------- **/
const double C1 = -4.0;				// for 12 Bit
const double C2 = +0.0405;			// for 12 Bit
const double C3 = -0.0000028;		// for 12 Bit
const double T1 = +0.01;			// for 14 Bit @ 5V
const double T2 = +0.00008;			// for 14 Bit @ 5V        
 
/** -------------------- SHT71 CMD -------------------- **/
#define STATUS_REG_W	0x06		// 000 0011 0
#define STATUS_REG_R	0x07		// 000 0011 1
#define MEASURE_TEMP	0x03		// 000 0001 1
#define MEASURE_HUMI	0x05		// 000 0010 1
#define RESET			0x1e		// 000 1111 0

#define ACK				1
#define	noACK			0

/** -------------------- C-Mapping ------------------- **/
#define	DATA_HI		DDRC  |=  (1<<PC0)
#define	DATA_LO		DDRC  &= ~(1<<PC0)

#define DATA_IN_HI	PINC & (1<<PINC0)
#define DATA_IN_LO	!(PINC & (1<<PINC0)

#define	SCK_HI		PORTC |=  (1<<PC1)
#define	SCK_LO		PORTC &= ~(1<<PC1)

unsigned int tempervalue[2]={0,0};

/** ------------------- SHT_write ------------------ **/
char SHT_write(unsigned char value)
{
	unsigned char i, error = 0;
	DDRC = 0b00000011;
	for (i = 0x80; i > 0; i /= 2)			
	{
		if (i & value)
		DATA_HI;						
		else DATA_LO;
		SCK_HI;								
		_delay_us(5);						
		SCK_LO;
	}
	DATA_HI;		
	DDRC = 0b00000010;
	SCK_HI;		
	error = PINC0;						
	_delay_us(2);
	SCK_LO;
	return error;				// error=1 in case of no acknowledge
}

/** ------------------- SHT_read ------------------- **/
char SHT_read(unsigned char ack)
{
	// reads a byte form the Sensibus and gives an acknowledge in case of "ack=1"
	unsigned char i,val=0;
	DDRC = 0b00000010;
	for (i=0x80;i>0;i/=2)					
	{ 	SCK_HI;
		_delay_us(2);
		if(PINC0) val=(val | i);
		_delay_us(2);
		SCK_LO;
	}
	DDRC = 0b00000011; 
	if(ack==1){DATA_LO;};
	SCK_HI;
	_delay_us(5);
	SCK_LO;
	DATA_HI;
	return val;
}

/** ------------------- SHT_transstart ------------------ **/
void SHT_transstart(void)
{
	//----------------------------------------------------------------------------------
	// generates a transmission start
	//    _____         ________
	// _DATA_: |_______|
	//          ___   ___
	// SCK : ___| |___| |______
	//----------------------------------------------------------------------------------
	DDRC = 0b00000011;
	DATA_HI; SCK_LO; //Initial state
	_delay_us(2);
	SCK_HI;
	_delay_us(2);
	DATA_LO;
	_delay_us(2);
	SCK_LO;
	_delay_us(5);
	SCK_HI;
	_delay_us(2);
	DATA_HI;
	_delay_us(2);
	SCK_LO;      
	DDRC = 0b00000010;
}

/** ---------------- SHT_connection_reset ---------------- **/
void SHT_connection_reset(void)
{
	//----------------------------------------------------------------------------------
	// communication reset: _DATA_-line=1 and at least 9 SCK cycles followed by transstart
	// _____________________________________________________         ________
	// _DATA_:                                              |_______|
	//          _    _    _    _    _    _    _    _    _       ___   ___
	// SCK : __| |__| |__| |__| |__| |__| |__| |__| |__| |______| |___| |______
	//----------------------------------------------------------------------------------
	unsigned char i;
	DDRC = 0b00000011;
	DATA_HI; SCK_LO;
	for(i = 0;i < 9; i++)
	{	SCK_HI; 
		_delay_us(1);
		SCK_LO;
		_delay_us(1);
	}
	SHT_transstart();
	DDRC = 0b00000010;
}

/** ------------------ SHT_softreset ------------------ **/
char SHT_softreset(void)
{
	// resets the sensor by a softreset
	unsigned char error=0;
	SHT_connection_reset();						// reset communication
	error+=SHT_write(RESET);					// send RESET-command to sensor
	return error;								// error=1 in case of no response form the sensor
}

/** ---------------- SHT_read_statusreg ---------------- **/
char SHT_read_statusreg(unsigned char *p_value, unsigned char *p_checksum)
{
	unsigned char error=0;
	SHT_transstart();							// transmission start
	error=SHT_write(STATUS_REG_R);				// send command to sensor
	*p_value=SHT_read(ACK);						// read status register (8-bit)
	*p_checksum=SHT_read(noACK);				// read checksum (8-bit)
	return error;								// error=1 in case of no response form the sensor
}

/** ---------------- SHT_write_statusreg --------------- **/
char SHT_write_statusreg(unsigned char *p_value)
{	
	unsigned char error=0;
	SHT_transstart();
	error+=SHT_write(STATUS_REG_W);
	error+=SHT_write(*p_value);
	return error;
}

/** ----------------- MEASUREMENT ----------------- **/
char SHT_measure(unsigned char *p_value, unsigned char *p_checksum, unsigned char mode)
{
	unsigned char error=0;
	char puffer[15];
	unsigned int temp=0;
	SHT_transstart();
	
	switch(mode)
	{		
		case TEMP : error+=SHT_write(MEASURE_TEMP); break;
		case HUMI : error+=SHT_write(MEASURE_HUMI); break;
		default: break;
	}
	DDRC = 0b00000010;
	while (1)
	{        
		if(PINC0 == 0) break;
	} 
	if(PINC0) error+=1;
	switch(mode){
	case TEMP		:	temp=0;
						temp=SHT_read(ACK);
						temp<<=8;
						tempervalue[0]=temp;
						sprintf(puffer,"tr=%d\n", temp);
						uart_puts(puffer);
						temp=0;
						temp=SHT_read(ACK);
						tempervalue[0]|=temp;
						
						break;
						
	case HUMI		: temp=0;
						temp=SHT_read(ACK);
						temp<<=8;
						tempervalue[1]=temp;
						temp=0;
						temp=SHT_read(ACK);
						tempervalue[1]|=temp;
						
						break;
						
	default			: break;         
	}   
	*p_checksum =SHT_read(noACK);  //read checksum     
	return error;
}	

/** ------------------- CALC SHT 71 ------------------- **/
double Calc_SHT71(double p_humidity ,double *p_temperature)
//-----------------------------------------------------
// calculates temperature [?] and humidity [%RH] 
// input :  humi [Ticks] (12 bit) 
//          temp [Ticks] (14 bit)
// output:  humi [%RH]
//          temp [?]
//-----------------------------------------------------
{	
	double rh_lin;							// rh_lin:  Humidity linear 
	double rh_true;							// rh_true: Temperature compensated humidity
	double t = *p_temperature;				// t:       Temperature [Ticks] 14 Bit 
	double rh = p_humidity;					// rh:      Humidity [Ticks] 12 Bit
	double t_C;								// t_C   :  Temperature [?]
	
	t_C = t * 0.01 - 40;					// calc. temperature from ticks to [?]
	rh_lin = C3 * rh * rh + C2 * rh + C1;	// calc. humidity from ticks to [%RH]
											// calc. temperature compensated humidity [%RH]
	rh_true = (t_C - 25) * (T1 + T2 * rh) + rh_lin;
	
	if(rh_true > 100)rh_true = 100;			// cut if the value is outside of
	if(rh_true < 0.1)rh_true = 0.1;			// the physical possible range
	
	*p_temperature = t_C;					// return temperature [?]
	
	return rh_true;
}

/** ----------------- Calc_DEWPOINT ------------------- **/
double Calc_dewpoint(double h,double t)
// calculates dew point
// input:   humidity [%RH], temperature [?]
// output:  dew point [?]
{
	double logEx,dew_point;
	logEx = 0.66077 + 7.5 * t / (237.3 + t) + (log10(h) - 2);
	dew_point = (logEx - 0.66077) * 237.3 / (0.66077 + 7.5 - logEx);
	return dew_point;
}

/** ------------------- SENSOR_CMD -------------------- **/
void SENSOR_CMD(void)
{
	value humi_val,temp_val;
	unsigned char error,checksum;
	double dewpoint=0.0;
	char inp;
	char puffer[40];

	SHT_connection_reset();
  	
	// Place your code here
	_delay_ms(500);        
	error=0;            
	error+=SHT_measure((unsigned char*)( &humi_val.i),&checksum,HUMI);  //measure humidity
	error+=SHT_measure((unsigned char*) (&temp_val.i),&checksum,TEMP);  //measure temperature               
	error += SHT_read_statusreg(&inp, &checksum);
	if(error!=0) 
	{
		SHT_connection_reset();                 //in case of an error: connection reset
		sprintf(puffer, "Error");
		uart_puts(puffer);   
	}       
	else
	{
		humi_val.f=(double)tempervalue[1];                   //converts integer to double
		temp_val.f=(double)tempervalue[0];                   //converts integer to double
		humi_val.f=Calc_SHT71(humi_val.f,&temp_val.f);      //calculate humidity, temperature
		dewpoint=Calc_dewpoint(humi_val.f,temp_val.f) ;
		
		sprintf(puffer,"rh= %d, tr=%d, ta=%d\n", (double)tempervalue[1], (double)tempervalue[0], (double)dewpoint);
		uart_puts(puffer);
	}
		
	_delay_ms(100);
	
}