/*
 *      uart.c
 *
 *      Copyright 2008 Ronny Schmiedel <http://avrs-at-leipzig.de>
 *
 *		Original from Peter Fleury <http://homepage.hispeed.ch/peterfleury/>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 *
 *      AVR306: Using the AVR UART in C using IRQ
 *      Port for AVR GCC/WinAVR
 *      Routines for interrupt controlled USART
 *      Supports ATMEGA8, ATMEGA16
 *      Last Edit: 02.01.2009
 */
 
 
/* Includes */
#include <avr/io.h>
#include <avr/interrupt.h>
#include "uart.h"
 
#define UBRR_BAUD ((F_CPU/(16L*UART_BAUD_RATE))-1)
 
/* Static Variables */
static volatile unsigned char UART_TxBuf[UART_TX_BUFFER_SIZE];
static volatile unsigned char UART_RxBuf[UART_RX_BUFFER_SIZE];
static volatile unsigned char UART_TxHead;
static volatile unsigned char UART_TxTail;
static volatile unsigned char UART_RxHead;
static volatile unsigned char UART_RxTail;
 
 
/* Initialize USART */
void uart_init(void)
{
       /* Special for ATMEGA8, ATMEGA16 */
       UCSRB = (( 1 << RXCIE ) | ( 1 << RXEN ) | ( 1 << TXEN ));   
	UCSRC = (1<<URSEL)|(1<<UCSZ1)|(1<<UCSZ0);                   
	UBRRH = (uint8_t) (UBRR_BAUD>>8);
	UBRRL = (uint8_t) (UBRR_BAUD & 0xFF);
 
 
	/* Flush receive buffer */
 
	UART_RxTail = 0;
	UART_RxHead = 0;
	UART_TxTail = 0;
	UART_TxHead = 0;
}
 
 
/* Interrupt handlers */
//*****Empfang der Daten*****
SIGNAL(SIG_UART_RECV)          /* change ISR-Name for another AVR */
{
	uint8_t data;
	uint8_t tmphead;
 
	/* Read the received data */
	data = UDR;
 			
	/* Calculate buffer index */
	tmphead = ( UART_RxHead + 1 ) & UART_RX_BUFFER_MASK;
 
	if ( tmphead == UART_RxTail )
	{
		/* ERROR! Receive buffer overflow */
	}
	else
	{
		        /* store new index */
        UART_RxHead = tmphead;
        /* store received data in buffer */
        UART_RxBuf[tmphead] = data;
        } 
}
 //********Senden der Daten****
SIGNAL(SIG_UART_DATA)   /* change ISR-Name for another AVR */ 
{
	uint8_t tmptail;
 
	/* Check if all data is transmitted */
	if ( UART_TxHead != UART_TxTail )
	{
		/* Calculate buffer index */
		tmptail = ( UART_TxTail + 1 ) & UART_TX_BUFFER_MASK;
		UART_TxTail = tmptail;      /* Store new index */
 
		UDR = UART_TxBuf[tmptail];  /* Start transmition */
	}
	else
	{      /* Disable UDRE interrupt, Special ATMEGA8, ATMEGA16 */
		UCSRB &= ~(1<<UDRIE);         
	}
}
 
/* Read and write functions */
uint8_t uart_getc( void )
{
	uint8_t tmptail;
	uint8_t data;
 
    /* calculate /store buffer index */
    tmptail = (UART_RxTail + 1) & UART_RX_BUFFER_MASK;
    UART_RxTail = tmptail; 
 
    /* get data from receive buffer */
    data = UART_RxBuf[tmptail];
 
	return data;
}
 
void uart_putc( uint8_t data )
{
 
	uint8_t tmphead;
 
 
    tmphead  = (UART_TxHead + 1) & UART_TX_BUFFER_MASK;
 
    while ( tmphead == UART_TxTail )
	{
        ;/* wait for free space in buffer */
    }
 
    UART_TxBuf[tmphead] = data;
    UART_TxHead = tmphead;
     /* Enable UDRE interrupt, special for ATMEGA8, ATMEGA16 */
	UCSRB |= (1<<UDRIE);                  
} /* uart_putc */
 
char uart_data_received( void )
{
    /* Return 0 (FALSE) if the receive buffer is empty */
	return ( UART_RxHead != UART_RxTail ); }
 
void uart_puts(const char *s )
{
    while (*s) 
      uart_putc(*s++);
 
}/* uart_puts */